/*

=head1 NAME

delicious.js - JavaScript interface to the del.icio.us API

=head1 SYNOPSIS

 <script type = "text/javascript" src = "delicious.js" />
 <script type = "text/javascript" src = "xmlp/xmlp.js" />

 # you'll need to do this under Mozilla/Gecko (but not Safari)        
 # netscape.security.PrivilegeManager.enablePrivilege("UniversalBrowserRead");

 var del = new Delicious({'username':'foo',
                          'password':'bar'});

 var posts = del.posts({'tag':'baz'});

 if (! posts) {
    alert(del.last_error());
    return false;
 }

 for (var i = 0; i < posts.length; i++) {
    var post = posts[i];

    alert(post.description() + ":" + post.href());
 }

=head1 DESCRIPTION

JavaScript interface to the del.icio.us API. 

This library is B<not> asynchronous. Additionally, it has built-in code
to throttle abusive requests : a mandatory pause of B<1> second between
requests to the del.icio.us server.

=head1 REQUIREMENTS

=over 4

=item * B<http://www.gazingus.org/html/XML_Parser_in_JavaScript.html>

For reasons unknown, some of the responses sent by the del.icio.us server
are not handled correctly by the I<XMLHttpRequest> object. Specifically,
the return value of the B<responseXML> method returns nothing.

Instead, we parse the value of the B<responseText> method by hand hence
the need for an XML parser.

=back

=cut

=head1 Delicious CLASS METHODS

=head1 new Delicious({key:value})

Valid arguments are :

=over 4

=item * B<username>

String; I<required>.

=item * B<password>

String; I<required>.

=item * B<retry>

Boolean.

If true, when a method encounters an HTTP 503 error from the del.icio.us
servers it will retry the request, not exceeding B<10> times.

The pause between each request will increase incrementally so, given
the non-asynchronous nature of the library, you should set this flag
carefully.

This flag does not affect the built-in throttling that occurs between
any other response/request cycle.

Default is false.

=back

Returns a Delicious object.

=cut

*/

function Delicious (args) {
    this.inheritsFrom = Base;
    this.inheritsFrom("Delicious");

    this._username = args['username'];
    this._password = args['password'];
    this._retry    = args['retry'];

    this._ua        = null;
    this._lasterror = null;

    // backoff-code variables

    this._wait                = 0;
    this._paused              = 0;
    this._maxtries            = 10;
    this._seconds_unavailable = 2;
    this._seconds_ok          = 1;

    this.VERSION = '1.0';
}

/* 

=head1 Delicious OBJECT METHODS

=head2 delicious.posts({key:value})

Get a list of posts on a given date, filtered by tag. If no date is
supplied, most recent date will be used.

Valid arguments are :

=over 4

=item * B<tag>

String.

Filter by this tag.

=item * B<dt>

String.

Filter by this date.

=back

Returns an array of I<Post> objects.

=cut

*/

Delicious.prototype.posts = function (args) {

    var optional = new Array("tag","dt");
    var params   = this._checkargs(args,null,optional);

    if (params == false) {
	return false;
    }

    var res = this._sendrequest('posts/get',params);
    return this._buildposts(res);
}

/*

=head2 delicious.posts_per_date({key:value})

Get a list of dates with the number of posts at each date.

Valid arguments are :

=over 4

=item * B<tag>

String.

Filter by this tag.

=back

Returns an array of I<Date> objects. B<This method is currently
not available.>

=cut

*/

Delicious.prototype.posts_per_date = function (args) {

    return this.error(500,"this method is not available");

    var optional = new Array("tag");
    var params   = this._checkargs(args,null,optional);

    if (params == false) {
	return false;
    }

    var res = this._sendrequest('posts/dates',params);
    return this._buildposts(res);
}

/*

=head2 delicious.recent_posts({key:value})

Get a list of most recent posts, possibly filtered by tag.

Valid arguments are :

=over 4

=item * B<tag>

String.

Filter by this tag.

=item * B<count>

Int.

Number of posts to return. Default is 20; maximum is 100.

=back

Returns an array of I<Post> objects.

=cut

*/

Delicious.prototype.recent_posts = function (args) {

    var optional = new Array("tag","count");
    var params   = this._checkargs(args,null,optional);

    if (params == false) {
	return false;
    }

    var res = this._sendrequest('posts/recent',params);
    return this._buildposts(res);
}

/*

=head2 delicious.all_posts()

Returns an array or I<Post> objects.

=cut

*/

Delicious.prototype.all_posts = function () {
    var res = this._sendrequest('posts/all');
    return this._buildposts(res);
}

/*

=head2 delicious.add_post(Array())

Makes a post to del.icio.us.

Valid arguments are :

=over 4

=item * B<url>

String; I<required>

=item * B<description>

String.

Description for post.

=item * B<extended>

String.

Extended description for post.

=item * B<tags>

String.

Space-delimited list of tags.

=item * B<dt>

String.

Datestamp for post, format "CCYY-MM-DDThh:mm:ssZ".

=back

Returns true or false.

=cut

*/

Delicious.prototype.add_post = function (args) {

    var required = new Array("url");
    var optional = new Array("description","extended","tags","dt");

    var params   = this._checkargs(args,required,optional);

    if (! params) {
	return false;
    }

    var res = this._sendrequest('posts/add',params);
    return this._isokay(res);
}

/*

=head2 delicious.delete_post({key:value})

Delete a post from del.icio.us.

Valid arguments are :

=over 4

=item * B<url>

String; I<required>.

=back

Returns true or false.

=cut

*/

Delicious.prototype.delete_post = function (args) {

    var required = new Array("url");
    var params   = this._checkargs(args,required);

    if (! params) {
	return false;
    }

    var res = this._sendrequest('posts/delete',params);
    return this._isokay(res);
}

/*

=head2 delicious.tags()

Returns an array of I<Tag> objects.

=cut

*/

Delicious.prototype.tags = function () {
    var res = this._sendrequest('tags/get');
    return this._buildtags(res);
}

/*

=head2 delicious.rename_tag({key:value})

Renames tags across all posts.

Valid arguments are :

=over 4

=item * B<old>

String; I<required>.

=item * B<new>

String; I<required>.

=back

Returns true or false.

=cut

*/

Delicious.prototype.rename_tag = function (args) {

    var required = new Array("old","new");
    var params   = this._checkargs(args,required);

    if (! params) {
	return false;
    }

    var res = this._sendrequest('tags/rename',params);
    return this._isokay(res);
}

/* 

=head2 delicious.last_error()

Returns an I<Error> object.

=cut

*/

Delicious.prototype.last_error = function () {
    return this._lasterror;
}

Delicious.prototype.error = function (code,errstr) {
    this._lasterror = new Error(code,errstr);
    return false;
}

Delicious.prototype._isokay = function (res) {
    if (! res) {
	return false;
    }

    var root = res.getDocumentElement();
    var code = null;

    /* Because 'rename_tag' will return
          <result>done</done>
       while the add|remove_post methods
       return : 
          <result code = "done" />
    */

    if (root.getAttribute('code')) {
	code = root.getAttribute('code');
    }

    else if (root.firstChild) {
	code = root.firstChild.nodeValue
    }

    else {
	return this.error(500,"unable to determine result code");
    }

    if (code != 'done') {
	return this.error(500,code);
    }

    return true;
}

Delicious.prototype._checkargs = function (args,required,optional) {
    var params = new Object ();

    if (! args) {
	return undefined;
    }

    if (required) {
	for (var i = 0; i < required.length; i++) {
	    var key = required[i];
	    
	    if (! args[key]) {
		return this.error(406,"missing required arg " + key);
	    }
	    
	    params[key] = args[key];
	}
    }

    if (optional) {
	for (var i = 0; i < optional.length; i++) {
	    var key = optional[i];
	    
	    if (args[key]) {
		params[key] = args[key];
	    }
	}
    }

    return params;
}

Delicious.prototype._buildposts = function (res) {

    if (! res) {
	return false;
    }

    var posts = res.getElementsByTagName("post");
    var list  = new Array ();

    for (var i = 0; i < posts.length; i++) {
	var attrs = posts[i].attributes;

	var args = new Array ();

	for (var j = 0; j < attrs.length; j++) {
	    var property = attrs.item(j);
	    args[property.nodeName] = property.nodeValue;
	}

	list[i] = new Post (args);
    }

    return list;
}

Delicious.prototype._buildtags = function (res) {

    if (! res) {
	return false;
    }

    var tags = res.getElementsByTagName("tag");
    var list = new Array ();

    for (var i = 0; i < tags.length; i++) {

	var tag = tags[i];
	
	list[i] = new Tag (tag.getAttribute("tag"),
			   tag.getAttribute("count"));
    }

    return list;
}

Delicious.prototype._parseresponse = function (txt) {
    var parser = null;

    try {
	parser = new XMLParser();
    } 

    catch (e) {
	return this.error(500,"");
    }

    try {
	parser.parse(txt);
	doc = parser.doc;
    }

    catch (e) {
	errstr = "failed to parse " + txt + ", " + e.message;
	return this.error(500,errstr);
    }

    return doc;
}

Delicious.prototype._sendrequest = function (method,args) {

    // check to see if we are running
    // wild and need to back off...

    var now = this._epoch();

    while (now < this._wait) {
	this._sleep(this._seconds_ok);
	now = this._epoch();
    }

    req = this._builduri(method,args);
    ua  = this._agent();

    if (! ua) {
	return this.error(500,"unable to create XMLHttpRequest object");
    }

    // alert(this._username+":"+this._password);

    ua.open("GET",req,false,this._username,this._password);
    ua.send(null);

    if (ua.status == 503) {

	if (! this._retry) {
	    var errstr = ua.statusText + " (" + req + ")";
	    return this.error(ua.status,errstr);
	}

	this._paused ++;

	if (this._pause > this._maxtries) {
	    var errstr = ua.statusText + " (" + req + "; " +
		         this._maxtries + "attempts)";
	    return this.error(ua.status,errstr);
	}

	this._sleep(this._seconds_unavailable * this._paused);
	return this._sendrequest(method,args);
    }

    this._wait   = this._epoch() + this._seconds_ok;
    this._paused = 0;

    if (ua.status == 200) {

	/* Because for some reason, the XMLHttpRequest
	   won't return a XML response for stuff like
	     <?xml ... ?><result code = "done" />
	   Who knows why...
	*/
	
	// alert(ua.responseText);
	return ua.responseXML || this._parseresponse(ua.responseText);
    }

    else {
	errstr = ua.statusText + " (" + req + ")";
	return this.error(ua.status,errstr);
    }

}

Delicious.prototype._builduri = function (method,args) {

    var uri = "http://del.icio.us/api/" + method;

    if ((arguments.length == 1) || (typeof(args) == "undefined")) {
	return uri;
    }

    query = new Array();

    for (key in args) {
	// alert("!" + key+":"+args[key]);
	query.push(key + "=" + args[key]);
    }

    return uri + "?" + query.join("&");
}

Delicious.prototype._agent = function () {

    if (this._ua != null) {
	return this._ua;
    }

    var xmlhttp = null;

    /*@cc_on @*/
    /*@if (@_jscript_version >= 5)
    // JScript gives us Conditional compilation, we can cope with old IE versions.
    // and security blocked creation of the objects.

    try {
        xmlhttp = new ActiveXObject("Msxml2.XMLHTTP");
    } 

    catch (e) {
        try {
	    xmlhttp = new ActiveXObject("Microsoft.XMLHTTP");
        } 

	catch (E) {
	    return null;
	}
    }

    @end @*/

    if (typeof(XMLHttpRequest) != 'undefined') {
	xmlhttp = new XMLHttpRequest();
    }

    if (xmlhttp != null) {
	this._ua = xmlhttp;
    }

    return xmlhttp;
}

Delicious.prototype._sleep = function (seconds) { 
    var then,now;

    then = this._epoch();
    now  = then;

    while((now - then) < seconds) {
	now = this._epoch();
    }
}

Delicious.prototype._epoch = function () {
    return parseInt(new Date().getTime() / 1000);
}

/*

=head1 Post OBJECT METHODS

A I<Post> object's toString() method is set up to return the
value of its B<href> method.

=cut

*/

function Post (args) {
    this.inheritFrom = Base;
    this.inheritFrom("Post");

    this._href        = null;
    this._description = null;
    this._extended    = null;
    this._tag         = null;
    this._time        = null;

    for (key in args) {
	var localkey = '_' + key;

	if (this.exists(localkey)) {
	    this[localkey] = args[key];
	}
    }

    this.isa("Post");
}

Post.prototype.toString = function () {
    return this.href();
}

/*

=head2 post.href()

Returns a string.

=cut

*/

Post.prototype.href = function () {
    return this._href;
}

/*

=head2 post.description()

Returns a string.

=cut

*/

Post.prototype.description = function () {
    return this._description;
}

/*

=head2 post.extended()

Returns a string.

=cut

*/

Post.prototype.extended = function () {
    return this._extended;
}

/*

=head2 post.tag()

Returns a string value of the tag list passed by the
del.icio.us server.

=cut

*/

Post.prototype.tag = function () {
    return this._tag;
}

/*

=head2 post.tags()

Returns the tag list as an array of I<Tag> objects.

=cut

*/

Post.prototype.tags = function () {
    var list = new Array ();

    for (tag in this.tag().split(" ")) {
	list.push(new Tag(tag));
    }

    return list;
}

/*

=head2 post.time()

Returns a string, formatted YYYY-MM-DD

=cut

*/

Post.prototype.time = function () {
    return this._time;
}

/* 

=head1 Tag OBJECT METHODS

A I<Tag> object's toString() method is set up to return the
value of the tag itself.

=cut

*/

function Tag (str,count) {
    this.inheritFrom = Base;
    this.inheritFrom("Tag");

    this._tagstr = str;
    this._count  = count;
}

Tag.prototype.toString = function () {
    return this._tagstr;
}

/*

=head2 tag.count() 

Returns an int.

=cut

*/

Tag.prototype.count = function () {
    return this._count;
}

/* 

=head1 Error OBJECT METHODS

An I<Error> object's toString() method is set up to return the
string concatenated of its B<code> and B<errstr> methods.

=cut

*/

function Error (code,str) {
    this.inheritFrom = Base;
    this.inheritFrom("Error");

    this._code = code;
    this._str  = str;
}

Error.prototype.toString = function () {
    return this.code() + " " + this.errstr();
}

/*

=head2 error.code()

Returns the error code reported by, or to, the object.

=cut

*/

Error.prototype.code = function () {
    return this._code;
}

/*

=head2 error.errstr()

Returns the optional error string reported by, or to, the
object.

=cut

*/

Error.prototype.errstr = function () {
    return this._str;
}

function Base (type) {
    this.inheritFrom = Object;
    this.inheritFrom();

    this._ima = type || typeof(this);
    
    this.isa = function () {
	return (type == this._ima);
    }

    this.exists = function (field) {
	if (field != null) {
	    for (var i in this) {
		
		if (i == field ) {
		    return true;
		}
	    }
	}

	return false;
    }
}

/*

=head1 VERSION 

1.0

=head1 DATE

$Date: 2004/11/17 21:45:51 $

=head1 AUTHOR

Aaron Straup Cope E<lt>http://www.aaronstraupcope.comE<gt>

=head1 SEE ALSO

http://del.icio.us/doc/api

=head1 LICENSE

Copyright (c) 2004 Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it under
the same terms as the Perl Artistic License.

=cut

$Id: delicious.js,v 1.12 2004/11/17 21:45:51 asc Exp $

*/
